module Main where

import HFANN
import Data.List
import Text.Printf

netFile, testFile, trainFile :: String
netFile = "zoo.ann"
trainFile = "zoo.data"
testFile = "scaled_zoo.data"

-----------------
--   Parsing   --
-----------------

parseLine :: String -> [Double]
parseLine s = map (\w -> (read w)::Double) (words s)

toDoubles :: [(String,String)] -> [([Double],[Double])]
toDoubles xs = map (\(a,b) -> (parseLine a, parseLine b)) xs

intoPairs :: [String] -> [(String,String)]
intoPairs [] = []
intoPairs (x:y:xs) = (x,y) : intoPairs xs

getTestTuples :: FilePath -> Int -> IO [([Double],[Double])]
getTestTuples path skip = do 
      fileData <- (readFile path)
      return $  toDoubles . intoPairs . (drop (2*skip+1)) $  (lines fileData)


-----------------
--   Display   --
-----------------

printN :: [Double] -> String
printN ns = (foldl (++) "" (map (\n -> printf " %6.3f " n) ns))

displayResults :: [[Double]] -> [([Double],[Double])] -> IO ()
displayResults ns ts = do
  mapM_ (\(n,(i,o)) -> putStrLn ((printN i) ++ " <- In | Out -> " ++ 
                                   (printN o) ++ " | NNet -> " ++ 
                                      (printN n))) (zip ns ts)

outputResults :: [[Double]] -> [([Double],[Double])] -> IO ()
outputResults ns ts = do
  writeFile "toplot.dat"  $ concat (map (\(n,(i,o)) -> 
                           (printN o) ++ (printN n) ++ "\n") (zip ns ts))

-------------------------------------------------------------------------
-- Note .. unScale and categorize are only good for single output nnets -
--    And only good for this problem with 7 categories                  -
-------------------------------------------------------------------------
unScale :: [Double] -> [Double]
unScale  = map (\d -> ((d * (6) + 8)/2))
 
categorize :: [Double] -> [Int]
categorize xs = 
    let cs = map (\d -> truncate (((7 * (d-1))/6) + 1)) xs
    in map (\c -> if (c == 8) then 7 else c) cs

hitRate :: [Int] -> [Int] -> (Int, Int)
hitRate x y =
   let hits = map (\(a,b) -> (if (a == b) then 1 else 0)) (zip x y)
   in
       (sum hits, length x)

showHits :: [[Double]] -> [([Double],[Double])] -> IO ()
showHits ns ts = do
   let pairs = map (\(n,(i,o)) -> (head n, head o)) (zip ns ts)
   let (c,d) = unzip pairs
   let ccat = categorize $ unScale c
   let dcat = categorize $ unScale d
   let summary = concat $ map (\(c,d) -> " " ++ 
                    show c ++ " " ++ show d ++ "\n" ) (zip ccat dcat)
   let (hits, total) = hitRate ccat dcat
     
   putStrLn $ "Summary\n" ++ summary
   putStrLn $ "Ratio of " ++ show hits ++ " out of " ++ show total ++ "\n"
   writeFile "toplot.dat" summary

calcMSE ::  [([Double],[Double])] -> [[Double]] -> Double
-- A bit of a hack since it only calcs MSE on one output
-- which is Ok here but if you have multiple outputs from
-- net it will need to be changed.
calcMSE ts nns = 
     (foldl (+) 0 (zipWith (\t n -> (abs (n - t))**2) testOuts nnOuts)) 
     / (fromInteger (toInteger (length nnOuts))) 
       where  testOuts = concat (snd (unzip ts))
              nnOuts   = concat nns

--------------
--   Main   --
--------------

main :: IO ()
main = do

  -- Read Training Data and Don't Use Training Section to Test
  fileData <- (readFile "scaled_zoo.data")
  hdr <- return $ head (take 1 (lines fileData))
  skipStr <-  return (head (words hdr))


  --- Select Only One Option Below -----

  --------------------------------------
  -------  Use unseen data only --------
  --------------------------------------
  skip <- return $ read skipStr

  --------------------------------------
  -------  Use entire dataset ----------
  --------------------------------------
  --skip <- return 0

  --------------------------------------

  -- Use only data that hasn't been seen in the training
  testTuples <- getTestTuples testFile skip
  putStrLn $ "Tesing Count -> " ++ show  (length testTuples)
  withSavedFann netFile $ \fann -> do

     -- Run test data into trained net
     nnetOut <- mapM (\x -> runFann fann x) $ fst $ unzip testTuples 

     displayResults nnetOut testTuples

     --outputResults nnetOut testTuples

     putStrLn $ "\nMSE -> " ++ show (calcMSE testTuples nnetOut) ++ "\n"

     showHits nnetOut testTuples